/***************************************************************************************
 * @file     bootloader.c
 * @version  V1.0
 * @date     10 September 2018
 *
 * @note
 * VORAGO Technologies
 *
 * @note
 * Copyright (c) 2013-2018 VORAGO Technologies.
 *
 * @par
 * BY DOWNLOADING, INSTALLING OR USING THIS SOFTWARE, YOU AGREE TO BE BOUND BY
 * ALL THE TERMS AND CONDITIONS OF THE VORAGO TECHNOLOGIES END USER LICENSE AGREEMENT.
 * THIS SOFTWARE IS PROVIDED "AS IS". NO WARRANTIES, WHETHER EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. VORAGO TECHNOLOGIES
 * SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 ****************************************************************************************/
 
/*****************************************************************************/ 
/* Include files                                                             */ 
/*****************************************************************************/

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>

#include "va108xx.h"
#include "uart.h"
#include "bootloader.h"
#include "xmodem.h"
#include "cyp15b102.h"
#include "m95m01.h"
#include "main.h"

/*****************************************************************************/ 
/* Local pre-processor symbols/macros ('#define')                            */ 
/*****************************************************************************/

#define CTRL0_STATUS_VAL ((uint32_t)(3+1)<<SPI_PERIPHERAL_CTRL0_SCRDV_Pos)

/*****************************************************************************/ 
/* Local type definitions ('typedef')                                        */ 
/*****************************************************************************/

/* Bootloader state machine states */
typedef enum en_bl_states
{
	ST_BEGIN = 0,
	ST_XMODEM_DL,
	ST_COPY_ROM,
	ST_RUN_APP,
	ST_READ_MEM,
	ST_ERROR
} en_bl_states_t;

/* Bootloader status/error codes */
typedef enum en_bl_status
{
	BL_SUCCESS = 0,
	BL_WAITING_FOR_CMD = 1,
	BL_WAITING_FOR_IMG = 2,
	BL_ERR_XMODEM = 3,
	BL_ERR_IMG_INVALID = 4,
	BL_ERR_VERIFY_FAIL = 5
} en_bl_status_t;

/*****************************************************************************/ 
/* Local variable definitions ('static')                                     */ 
/*****************************************************************************/

/*****************************************************************************/ 
/* Local function prototypes ('static')                                      */ 
/*****************************************************************************/

#ifdef REB1_M95M01
static bool CopyApplicationImageToEEPROM(uint32_t baseAddress, uint32_t numBytes);
#else
static bool CopyApplicationImageToFRAM(uint32_t baseAddress, uint32_t numBytes);
#endif
static void ReadMem(uint32_t baseAddress, uint32_t endAddress);

/*****************************************************************************/ 
/* Function implementation - global ('extern') and local ('static')          */ 
/*****************************************************************************/

/*******************************************************************************
 **
 ** @brief  Checks the application FW image and returns true if it is valid
 **
 ******************************************************************************/
bool Bootloader_CheckAppIsValid(void)
{
	/* get the application reset vector */
	uint32_t resetVector = *((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET));
	
	/* application reset vector must point to within application space */
	if((resetVector >= (APPLICATION_START_ADDR+VECTOR_TABLE_LEN)) && (resetVector < APPLICATION_END_ADDR)) 
	{
		/* reset vector is an acceptable value */
		return true;
	}
	return false;
}

/*******************************************************************************
 **
 ** @brief  Performs cleanup, copies the IVT, and runs application image
 **
 ******************************************************************************/
void Bootloader_RunApp(void)
{
	/* Feed watchdog. */ 
	WDFeed();
	
	/* Clear all interrupts set. */
	NVIC->ICER[0] = 0xFFFFFFFF;
	NVIC->ICPR[0] = 0xFFFFFFFF;
	
	/* Allow writes to code memory space */
	VOR_SYSCONFIG->ROM_PROT |= SYSCONFIG_ROM_PROT_WREN_Msk;
	
	/* Copy vector table to Cortex-M0 fixed IVT space (0x0)*/
	/* This throws a null warning because we are copying to the start of Cortex M0 memory space which is 0x0 */
	memcpy((void*)(BOOTLOADER_START_ADDR+VECTOR_TABLE_OFFSET),(void*)(APPLICATION_START_ADDR+VECTOR_TABLE_OFFSET),VECTOR_TABLE_LEN);
	
	/* Disable re-loading from FRAM/code ROM on soft reset */
	VOR_SYSCONFIG->RST_CNTL_ROM &= ~SYSCONFIG_RST_CNTL_ROM_SYSRSTREQ_Msk;
	
	/* Generate a soft reset */
	NVIC_SystemReset();
}

/*******************************************************************************
 **
 ** @brief  Bootloader state machine
 **
 ******************************************************************************/
void Bootloader_Task(void)
{
	static en_bl_status_t status = BL_WAITING_FOR_CMD;
	static en_bl_states_t state = ST_BEGIN;
	static size_t bytesRcvd = 0;
	uint8_t rxByte;
#ifdef PC_MODE
#define UART_BUF_SZ 16
	char strBufTx[UART_BUF_SZ];
	static bool request_cmd = true;
#endif
	
	switch(state)
	{
		case ST_BEGIN:
		  /* wait for bootloader command (upload, verify, read, etc) */
		  status = BL_WAITING_FOR_CMD;
#ifdef PC_MODE
      if (true == request_cmd)
			{
				/* print help */
				request_cmd = false;
				UART0TxStr("\r\nSend cmd: u = upload, m = read memory, s = get status, r = reset\r\n");
			}  
#endif
			if (UART0IsRxByteAvailable())
	    {
				rxByte = UART0RxByte();
			  if('u' == rxByte)
			  {
					/* upload new user code */
					state = ST_XMODEM_DL;
					status = BL_WAITING_FOR_IMG;
			  }
				if('r' == rxByte)
				{
					/* reset */
					NVIC_SystemReset();
				}
				if ('s' == rxByte)
				{
					/* status request */
#ifdef PC_MODE
					if (Bootloader_CheckAppIsValid() == true)
					{
					  UART0TxStr("\r\nUser image valid");
					}
					else
					{
						UART0TxStr("\r\nUser image invalid - waiting for image upload");
					}
					request_cmd = true;
#else
					if (Bootloader_CheckAppIsValid() == true)
					{
					  UART0TxByte((uint8_t)BL_SUCCESS);
					}
					else
					{					
            UART0TxByte((uint8_t)BL_ERR_IMG_INVALID);
					}
#endif
				}
				if ('m' == rxByte)
				{
					state = ST_READ_MEM;
					status = BL_SUCCESS;
				}
	    }
		  break;
		
		case ST_XMODEM_DL:
#ifdef PC_MODE
			UART0TxStr("\r\nUpload application .bin using XMODEM-CRC\r\n");
#endif
			bytesRcvd = XMODEM_ReceiveFile((uint8_t*)APPLICATION_START_ADDR, (uint8_t*)APPLICATION_END_ADDR);
			if (bytesRcvd > 0)
			{
				/* Transfer complete */
#ifdef PC_MODE
				waitMs(100); 
			  UART0TxStr("\r\nTransfer complete, copying to boot ROM\r\n");
#endif
		    state = ST_COPY_ROM;
				status = BL_SUCCESS;
			}
			else 
			{
				state = ST_ERROR;
				status = BL_ERR_XMODEM;
			}
		  break;
			
		case ST_COPY_ROM:
		  /* copy image from RAM to FRAM (or EEPROM if REB1 eval board),
		     and verify a successful burn. If successful, write the reset vector
		     to flag the application image as valid/bootable */
#ifdef REB1_M95M01
  		if (CopyApplicationImageToEEPROM(APPLICATION_START_ADDR, bytesRcvd) == true)
#else
			if (CopyApplicationImageToFRAM(APPLICATION_START_ADDR, bytesRcvd) == true)
#endif
			{
#ifdef PC_MODE
			  UART0TxStr("Write and verify complete, running application\r\n");
				waitMs(10);
#endif
				state = ST_RUN_APP;
				status = BL_SUCCESS;
			}
			else
			{
				state = ST_ERROR;
				status = BL_ERR_VERIFY_FAIL;
			}
			break;
			
		case ST_RUN_APP:
			/* run the application, first checking if it is valid/safe to run */
		  if (Bootloader_CheckAppIsValid())
			{
				/* this function will never return - copies app IVT then generates a soft reset
				   if it somehow does return, throw an error and restart */
				Bootloader_RunApp();
			}	
			state = ST_ERROR;
			status = BL_ERR_IMG_INVALID;
			break;
			
		case ST_READ_MEM:
			/* read back user code space */
		  ReadMem(APPLICATION_START_ADDR, APPLICATION_END_ADDR);
		  state = ST_BEGIN;
		  status = BL_SUCCESS;
#ifdef PC_MODE	
			/* print the help menu again next time around */
		  request_cmd = true;
#endif
		  break;
			
		case ST_ERROR:
			/* Error - restart upload sequence */
#ifdef PC_MODE
		  sprintf(strBufTx, "ERROR %u\r\n", status);
			UART0TxStr(strBufTx);
		  request_cmd = true;
		  state = ST_BEGIN;
#else
			/* wait until status code is read */
			if (UART0IsRxByteAvailable())
	    {
			  if(UART0RxByte() == 's')
			  {
					/* Send status (error code) */
					UART0TxByte((uint8_t)status);
					state = ST_BEGIN;
			  }
	    }
#endif
		  break;
		
		default:
			/* How did it get here? throw an error */
			state = ST_ERROR;
			return;
	}
}

/*******************************************************************************
 **
 ** @brief  Writes the application FW image space in RAM over to FRAM
 ** Returns true if the copy was successful
 **
 ******************************************************************************/
#ifndef REB1_M95M01
static bool CopyApplicationImageToFRAM(uint32_t baseAddress, uint32_t numBytes)
{
	uint32_t addr;
	uint32_t bytesToWrite = numBytes;
	uint32_t resetVector;
	uint32_t verify;
	
	if (baseAddress < APPLICATION_START_ADDR)
	{
	  /* attempt to overwrite bootloader space - DO NOT DO THIS */
		return false;
	}
	if (baseAddress + bytesToWrite > APPLICATION_END_ADDR)
	{
	  /* attempt to write outside of application space - ALSO BAD */
		return false;
	}
	if ((bytesToWrite == 0) || ((bytesToWrite % 4) != 0))
	{
		/* nothing to copy, or not 32-bit word aligned */
		return false;
	}
	addr = baseAddress;
	
	/* Addresses look good, lets do it
	   We want to clear the application reset vector first in RAM, 
	   write the whole image from RAM, then write the reset vector 
	   last to ensure the app image will be flagged as invalid if
	   the write procedure fails or is interrupted by a reset */
	
	/* save the reset vector */
	resetVector = *((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET));
	
	/* now clear it in program RAM */
	VOR_SYSCONFIG->ROM_PROT |= SYSCONFIG_ROM_PROT_WREN_Msk; /* Allow writes to code memory space */
	*((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET)) = 0xFFFFFFFF; /* an invalid address */
	
	if(Bootloader_CheckAppIsValid() == true)
	{
		/* something went wrong - AppValid should now be false because reset vector cleared - ABORT */
		return false;
	}
	
	CYP15B102_Init(baseAddress,2000,2); /* these values aren't used so they don't really matter */
	while(bytesToWrite >= CYP15B102_SZPAGE)
	{
    CYP15B102_ProgramPage(addr, CYP15B102_SZPAGE, (uint8_t*)addr);
		bytesToWrite -= CYP15B102_SZPAGE;
		addr += CYP15B102_SZPAGE;
		
		/* feed watchdog */
		WDFeed();
	}
	if (bytesToWrite > 0)
	{
		/* write the last page if the end of code doesn't fall on a page boundary */
		CYP15B102_ProgramPage(addr, bytesToWrite, (uint8_t*)addr);
		
		/* feed watchdog */
		WDFeed();
	}
	
	/* verify the image */
	verify = CYP15B102_Verify(baseAddress, numBytes, (uint8_t*)baseAddress);
	if (verify != (baseAddress + numBytes))
	{
		/* verify now contains the address of first mismatch */
	
		/* Finished writing to EEPROM */
		CYP15B102_UnInit(2);
		
		/* indicate that verify failed */
		return false;
	}
	
	/* If the verify was successful, now write the reset vector to flag the image as valid */
	CYP15B102_ProgramPage(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET, 4, (uint8_t*)&resetVector);
	
	/* Finished writing to EEPROM */
	CYP15B102_UnInit(2);
	
	/* restore reset vector in program RAM */
	*((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET)) = resetVector; 
	VOR_SYSCONFIG->ROM_PROT &= ~SYSCONFIG_ROM_PROT_WREN_Msk; /* Disable writes to code memory space */
	
	if(Bootloader_CheckAppIsValid() == true)
	{
		/* Reset vector correctly restored */
    return true;
	}
	return false;
}
#endif

/*******************************************************************************
 **
 ** @brief  Writes the application FW image space in RAM over to EEPROM
 ** Returns true if the copy was successful
 **
 ** @param  baseAddress
 ** The starting address - usually the beginning of user application space
 ** 
 ** @param  numBytes
 ** Number of bytes to write, starting from baseAddress
 **
 ** @return true if write successful, false if write or verification failed
 **
 ******************************************************************************/
#ifdef REB1_M95M01
static bool CopyApplicationImageToEEPROM(uint32_t baseAddress, uint32_t numBytes)
{
	uint32_t addr;
	uint32_t bytesToWrite = numBytes;
	uint32_t resetVector;
	uint32_t verify;
	
	if (baseAddress < APPLICATION_START_ADDR)
	{
	  // attempt to overwrite bootloader space - DO NOT DO THIS
		return false;
	}
	if (baseAddress + bytesToWrite > APPLICATION_END_ADDR)
	{
	  // attempt to write outside of application space - ALSO BAD
		return false;
	}
	if (bytesToWrite == 0)
	{
		// nothing to copy
		return false;
	}
	addr = baseAddress;
	
	/* Addresses look good, lets do it
	   We want to clear the application reset vector first in RAM, 
	   write the whole image from RAM, then write the reset vector 
	   last to ensure the app image will be flagged as invalid if
	   the write procedure fails or is interrupted by a reset */
	
	// save the reset vector
	resetVector = *((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET));
	
	// now clear it in program RAM
	VOR_SYSCONFIG->ROM_PROT |= SYSCONFIG_ROM_PROT_WREN_Msk; // Allow writes to code memory space
	*((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET)) = 0xFFFFFFFF; // an invalid address
	
	if(Bootloader_CheckAppIsValid() == true)
	{
		// something went wrong - AppValid should now be false because reset vector cleared - ABORT
		return false;
	}
	
	M95M01_Init(baseAddress,2000,2); // these values aren't used so they don't really matter
	while(bytesToWrite >= M95M01_SZPAGE)
	{
    M95M01_ProgramPage(addr, M95M01_SZPAGE, (uint8_t*)addr);
		bytesToWrite -= M95M01_SZPAGE;
		addr += M95M01_SZPAGE;
		
		// feed watchdog
		WDFeed();
	}
	if (bytesToWrite > 0)
	{
		// write the last page if the end of code doesn't fall on a page boundary 
		M95M01_ProgramPage(addr, bytesToWrite, (uint8_t*)addr);
		
		// feed watchdog
		WDFeed();
	}
	
	// verify the image
	verify = M95M01_Verify(baseAddress, numBytes, (uint8_t*)baseAddress);
	if (verify != (baseAddress + numBytes))
	{
		// verify now contains the address of first mismatch
	
		// Finished writing to EEPROM
		M95M01_UnInit(2);
		
		// indicate that verify failed
		return false;
	}
	
	// If the verify was successful, now write the reset vector to flag the image as valid
	M95M01_ProgramPage(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET, 4, (uint8_t*)&resetVector);
	
	// Finished writing to EEPROM
	M95M01_UnInit(2);
	
	// restore reset vector in program RAM 
	*((uint32_t*)(APPLICATION_START_ADDR+RESET_VECTOR_OFFSET)) = resetVector; 
	VOR_SYSCONFIG->ROM_PROT &= ~SYSCONFIG_ROM_PROT_WREN_Msk; /* Disable writes to code memory space */
	
	if(Bootloader_CheckAppIsValid() == true)
	{
		// Reset vector correctly restored
    return true;
	}
	return false;
}
#endif

 /*******************************************************************************
 **
 ** @brief  Read user code space from application RAM, send it across the
 **         UART via XMODEM
 **
 ** @param  baseAddress
 **         The starting address
 **
 ** @param  endAddress
 **         The end address
 **
 ** @return none
 **
 ******************************************************************************/
static void ReadMem(uint32_t baseAddress, uint32_t endAddress)
{
	XMODEM_SendFile((uint8_t*)baseAddress, (size_t)(endAddress - baseAddress));
	return;
}

/*****************************************************************************/ 
/* End of file                                                               */ 
/*****************************************************************************/
