/***************************************************************************************
 * @file     uart.c
 * @version  V1.0
 * @date     18. January 2018
 *
 * @note
 * VORAGO Technologies 
 *
 * @note
 * Copyright (c) 2013-2016 VORAGO Technologies. 
 *
 * @par
 * BY DOWNLOADING, INSTALLING OR USING THIS SOFTWARE, YOU AGREE TO BE BOUND BY 
 * ALL THE TERMS AND CONDITIONS OF THE VORAGO TECHNOLOGIES END USER LICENSE AGREEMENT. 
 * THIS SOFTWARE IS PROVIDED "AS IS". NO WARRANTIES, WHETHER EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. VORAGO TECHNOLOGIES 
 * SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL, OR CONSEQUENTIAL 
 * DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 ****************************************************************************************/
/** @file */ 
/** \addtogroup UART
 *  @{
 */

#include "uart.h"
#include "main.h"

/* Variables for storing characters received from UART0. */
static uint8_t uart0RxBuffer[configUART_RX_BUFFER_SIZE];
static uint8_t *uart0RxBufferHead;
static uint8_t *uart0RxBufferTail;
static uint8_t *uart0RxBufferEnd;

/* Variables for storing characters received from UART1. */
static uint8_t uart1RxBuffer[configUART_RX_BUFFER_SIZE];
static uint8_t *uart1RxBufferHead;
static uint8_t *uart1RxBufferTail;
static uint8_t *uart1RxBufferEnd;

/*******************************************************************************
 **
 ** @brief Configures UART0 peripheral.
 **
 **  Initializes buffers and configures UART  
 **
 ******************************************************************************/
void UART0Init(void)
{
	 VOR_SYSCONFIG->PERIPHERAL_CLK_ENABLE |= CLK_ENABLE_UARTA;
	
    /* Initialize variables. */
    uart0RxBufferHead = uart0RxBuffer;
    uart0RxBufferTail = uart0RxBuffer;
    uart0RxBufferEnd = uart0RxBuffer + configUART_RX_BUFFER_SIZE;

    /* Configure UART to operate at configUART_BAUDRATE. */
    VOR_UART0.CLKSCALE = BaudToUARTCLKSCALE(configUART_BAUDRATE);
                     
    /* Configure word size and RTS behavior. */
    VOR_UART0.CTRL = (3 << UART_PERIPHERAL_CTRL_WORDSIZE_Pos) |
                      (UART_PERIPHERAL_CTRL_DEFRTS_Msk);

    /* Enable CTS flow control IO, if needed. */
#ifdef configUART_CTS_FLOW_CONTROL
    VOR_UART0.CTRL |= UART_PERIPHERAL_CTRL_AUTOCTS_Msk;
#endif

    /* Enable RTS flow control IO, if needed. */
#ifdef configUART_RTS_FLOW_CONTROL
    VOR_UART0.CTRL |= UART_PERIPHERAL_CTRL_AUTORTS_Msk;
#endif

	/* Enable RX interupts as soon as a character is received. */
	VOR_UART0.IRQ_ENB = UARTA_IRQ_ENB_IRQ_RX_Msk;
	VOR_UART0.RXFIFOIRQTRG = 1;
    
	VOR_IRQSEL->UART[0] = VOR_UART0_IRQ_INDEX;
	NVIC_SetPriority(VOR_UART0_IRQ_INDEX, 1);
	NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);	

    /* Enable UART. */
  VOR_UART0.ENABLE = (UART_PERIPHERAL_ENABLE_RXENABLE_Msk | 
                      UART_PERIPHERAL_ENABLE_TXENABLE_Msk);
	//send a break to let rx state machine reset										
	UART0SendBreak(32);
}


/*******************************************************************************
 **
 ** @brief Configures UART1 peripheral.
 **
 **  Initializes buffers and configures UART  
 **
 ******************************************************************************/
void UART1Init(void)
{
	 VOR_SYSCONFIG->PERIPHERAL_CLK_ENABLE |= CLK_ENABLE_UARTB;
	
    /* Initialize variables. */
    uart1RxBufferHead = uart1RxBuffer;
    uart1RxBufferTail = uart1RxBuffer;
    uart1RxBufferEnd = uart1RxBuffer + configUART_RX_BUFFER_SIZE;

    /* Configure UART to operate at configUART_BAUDRATE. */
    VOR_UART1.CLKSCALE = BaudToUARTCLKSCALE(configUART_BAUDRATE);
                     
    /* Configure word size and RTS behavior. */
    VOR_UART1.CTRL = (3 << UART_PERIPHERAL_CTRL_WORDSIZE_Pos) |
                      (UART_PERIPHERAL_CTRL_DEFRTS_Msk);

    /* Enable CTS flow control IO, if needed. */
#ifdef configUART_CTS_FLOW_CONTROL
    VOR_UART1.CTRL |= UART_PERIPHERAL_CTRL_AUTOCTS_Msk;
#endif

    /* Enable RTS flow control IO, if needed. */
#ifdef configUART_RTS_FLOW_CONTROL
    VOR_UART1.CTRL |= UART_PERIPHERAL_CTRL_AUTORTS_Msk;
#endif

	/* Enable RX interupts as soon as a character is received. */
	VOR_UART1.IRQ_ENB = UARTB_IRQ_ENB_IRQ_RX_Msk;
	VOR_UART1.RXFIFOIRQTRG = 1;
    
	VOR_IRQSEL->UART[1] = VOR_UART1_IRQ_INDEX;
	NVIC_SetPriority(VOR_UART1_IRQ_INDEX, 1);
	NVIC_EnableIRQ(VOR_UART1_IRQ_INDEX);	

    /* Enable UART. */
  VOR_UART1.ENABLE = (UART_PERIPHERAL_ENABLE_RXENABLE_Msk | 
                      UART_PERIPHERAL_ENABLE_TXENABLE_Msk);
	//send a break to let rx state machine reset										
	//UART1SendBreak(32);
}

/*******************************************************************************
 **
 ** @brief  Transmit a string across UART0.
 **
 ** Transmits a string across UART0, blocking until all characters are sent.
 **
 ** @param  pcStr is the pointer to the string that is transmitted across UART0.
 **
 ******************************************************************************/
void UART0TxStr(char *pcStr)
{
	while(*pcStr)
    {
        /* Block until UART is done transmitting data. */
        while((VOR_UART0.TXSTATUS & UART_PERIPHERAL_TXSTATUS_WRRDY_Msk) == 0)
        {
        }
        
        /* Transmit another character. */
        VOR_UART0.DATA = *pcStr++;
	}
}

/*******************************************************************************
 **
 ** @brief  TX one byte over UART0
 **
 ** @param  uint8_t b is the byte to send
 **
 ******************************************************************************/
void UART0TxByte(uint8_t b)
{
	/* Block until UART is done transmitting data. */
  while((VOR_UART0.TXSTATUS & UART_PERIPHERAL_TXSTATUS_WRRDY_Msk) == 0)
  {
  }
	VOR_UART0.DATA = b;
}

/*******************************************************************************
 **
 ** @brief  RX one byte over UART0, wait for a byte but timeout if no byte is received
 **
 ******************************************************************************/
uint8_t UART0RxByte(void)
{
	uint32_t timer = 1000000;
	uint8_t b =0;
	while(--timer)
	{
		if(UART0Read(&b))
		{
			break;
		}
	}
	return b;
}

/*******************************************************************************
 **
 ** @brief  Returns true if a byte is available on the RX buffer, else false
 **
 ******************************************************************************/
bool UART0IsRxByteAvailable(void)
{
	/* Disable UART RX interrupt. */
  NVIC_DisableIRQ(VOR_UART0_IRQ_INDEX);
    
  /* Verify that data is available. */
  if(uart0RxBufferTail == uart0RxBufferHead)
  {
		/* Re-enable UART RX interrupt. */
    NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);
		return false;
	}
	/* Re-enable UART RX interrupt. */
  NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);
	return true;
}

/*******************************************************************************
 **
 ** @brief  Transmit a string across UART1.
 **
 ** Transmits a string across UART1, blocking until all characters are sent.
 **
 ** @param  pcStr is the pointer to the string that is transmitted across UART1.
 **
 ******************************************************************************/
void UART1TxStr(char *pcStr)
{
	while(*pcStr)
    {
        /* Block until UART is done transmitting data. */
        while((VOR_UART1.TXSTATUS & UART_PERIPHERAL_TXSTATUS_WRRDY_Msk) == 0)
        {
        }
        
        /* Transmit another character. */
        VOR_UART1.DATA = *pcStr++;
	}
}

/*******************************************************************************
 **
 ** @brief  TX one byte over UART1
 **
 ** @param  uint8_t b is the byte to send
 **
 ******************************************************************************/
void UART1TxByte(uint8_t b)
{
	/* Block until UART is done transmitting data. */
  while((VOR_UART1.TXSTATUS & UART_PERIPHERAL_TXSTATUS_WRRDY_Msk) == 0)
  {
  }
	VOR_UART1.DATA = b;
}


/*******************************************************************************
 **
 ** @brief  Reads a character from the UART0 RX buffer.
 **
 ** @param  data is the pointer to store received data, if any exists.
 **
 ** @return Returns **true** if data was pulled from the RX buffer, otherwise
 **          returns **false** and zeros out data.
 **
 ******************************************************************************/
bool UART0Read(uint8_t *data)
{   
    /* Disable UART RX interrupt. */
    NVIC_DisableIRQ(VOR_UART0_IRQ_INDEX);
    
    /* Verify that data is available. */
    if(uart0RxBufferTail == uart0RxBufferHead)
    {
        /* Zero out data pointer. */
        *data = 0;
        
        /* Re-enable UART RX interrupt. */
        NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);
        
        return(false);
    }
    
    /* Pull byte from buffer. */
	*data = *uart0RxBufferTail++;
    
    /* Loop pointer if needed. */
	if(uart0RxBufferTail == uart0RxBufferEnd)
    {
		uart0RxBufferTail = uart0RxBuffer;
    }
    
    /* Re-enable UART RX interrupt. */
	NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);
    
  return(true);
}

/*******************************************************************************
 **
 ** @brief  Reads from from the UART0 RX buffer
 **
 ** @param  data is the pointer to store received data, if any exists.
 **         limit prevents coping enough to overrun buffer in calling enviro
 **
 ** @return Returns num of char copied out.  Data may or may not not be 
 **         of string type
 **
 ******************************************************************************/
uint32_t UART0GetBuffer(uint8_t *data, uint32_t limit)
{   
	uint32_t count=0;
	/* Disable UART RX interrupt. */
	NVIC_DisableIRQ(VOR_UART0_IRQ_INDEX);
	
	/* Verify that data is available. */
	while((uart0RxBufferTail != uart0RxBufferHead) & limit)
	{
		++count;
		--limit;
		/* Pull byte from buffer. */
		*data++ = *uart0RxBufferTail++;
		
	 /* Loop pointer if needed. */
		if(uart0RxBufferTail == uart0RxBufferEnd)
		{
			uart0RxBufferTail = uart0RxBuffer;
		}
	}
    
  /* Re-enable UART RX interrupt. */
	NVIC_EnableIRQ(VOR_UART0_IRQ_INDEX);
    
  return(count);
}

/*******************************************************************************
 **
 ** @brief  Reads from from the UART1 RX buffer
 **
 ** @param  data is the pointer to store received data, if any exists.
 **         limit prevents coping enough to overrun buffer in calling enviro
 **
 ** @return Returns num of char copied out.  Data may or may not not be 
 **         of string type
 **
 ******************************************************************************/
uint32_t UART1GetBuffer(uint8_t *data, uint32_t limit)
{   
	uint32_t count=0;
	/* Disable UART RX interrupt. */
	NVIC_DisableIRQ(VOR_UART1_IRQ_INDEX);
	
	/* Verify that data is available. */
	while((uart1RxBufferTail != uart1RxBufferHead) & limit)
	{
		++count;
		--limit;
		/* Pull byte from buffer. */
		*data++ = *uart1RxBufferTail++;
		
	 /* Loop pointer if needed. */
		if(uart1RxBufferTail == uart1RxBufferEnd)
		{
			uart1RxBufferTail = uart1RxBuffer;
		}
	}
    
  /* Re-enable UART RX interrupt. */
	NVIC_EnableIRQ(VOR_UART1_IRQ_INDEX);
    
  return(count);
}


/*******************************************************************************
 **
 ** @brief  UART0 RX interrupt handler.
 **
 ** Reads characters received into RX buffer.  If the RX buffer is full, the
 ** contents of the RX buffer will be overwritten.
 **
 ******************************************************************************/
void VOR_UART0_IRQHandler(void)
{
    /* Read bytes into buffer as long as bytes are available. */
	while((VOR_UART0.RXSTATUS & UART_PERIPHERAL_RXSTATUS_RDAVL_Msk) != 0)
    {
		*uart0RxBufferHead++ = VOR_UART0.DATA;
		
        /* Loop pointer if needed. */
		if(uart0RxBufferHead == uart0RxBufferEnd)
        {
            uart0RxBufferHead = uart0RxBuffer;
        }
	}
}

/*******************************************************************************
 **
 ** @brief  UART1 RX interrupt handler.
 **
 ** Reads characters received into RX buffer.  If the RX buffer is full, the
 ** contents of the RX buffer will be overwritten.
 **
 ******************************************************************************/
void VOR_UART1_IRQHandler(void)
{
    /* Read bytes into buffer as long as bytes are available. */
	while((VOR_UART1.RXSTATUS & UART_PERIPHERAL_RXSTATUS_RDAVL_Msk) != 0)
    {
		*uart1RxBufferHead++ = VOR_UART1.DATA;
		
        /* Loop pointer if needed. */
		if(uart1RxBufferHead == uart1RxBufferEnd)
        {
            uart1RxBufferHead = uart1RxBuffer;
        }
	}
}

/*****************************************************************************/ 
/* End of file                                                               */ 
/*****************************************************************************/
